@@ -67,7 +67,7 @@
   (when xcb:debug-on
     `(message (concat "[XELB LOG] " ,format-string) ,@args)))
 
-(defvar xcb:connection-timeout 10 "Connection timeout.")
+(defvar xcb:connection-timeout 3 "Connection timeout.")
 
 ;;;; X connection related
 
@@ -91,24 +91,11 @@
    (extension-first-error-alist :initform nil)
    (extension-first-event-alist :initform nil)
    (request-sequence :initform 0)
-   (error-sequence :initform 0)
-   (reply-sequence :initform 0)
+   (last-seen-sequence :initform 0)
    (xid :initform 0)            ;last used X resource ID
    (extra-plist :initform nil)) ;for storing extra data (e.g. by extensions)
   :documentation "X connection.")
 
-(defsubst xcb:-sequence-cmp16 (sequence1 sequence2)
-  "Compare 16-bit sequence numbers SEQUENCE1 and SEQUENCE2.
-
-Return a positive value if SEQUENCE1 is larger than SEQUENCE2, 0 if they are
-equal.  Otherwise a negative value would be returned."
-  (if (= sequence1 sequence2)
-      0
-    (let ((diff (- sequence1 sequence2)))
-      (if (< #x7FFF (abs diff))
-          (- diff)                      ;overflowed
-        diff))))
-
 (defclass xcb:auth-info ()
   ((name :initarg :name :initform "" :type string)
    (data :initarg :data :initform "" :type string))
@@ -213,7 +200,7 @@ equal.  Otherwise a negative value would be returned."
     (set-process-plist process
                        (plist-put (process-plist process) 'connection obj))
     (set-process-coding-system process 'binary 'binary)
-    (set-process-filter process 'xcb:-connection-filter)
+    (set-process-filter process #'xcb:-connection-setup-filter)
     (process-send-string                ;send setup packet
      process
      (apply #'unibyte-string
@@ -234,6 +221,66 @@ equal.  Otherwise a negative value would be returned."
       (while (not (slot-value obj 'setup-data))
         (accept-process-output process 1 nil 1)))))
 
+(defconst xcb:-SEQUENCE-SEGMENT-MASK (lognot #xFFFF))
+
+(defun xcb:-connection-setup-filter (process message)
+  "Process filter used during connection setup."
+  (let* ((connection (plist-get (process-plist process) 'connection))
+         (cache (vconcat (slot-value connection 'message-cache) message)))
+    (setf (slot-value connection 'message-cache) cache)
+    (unless (or (slot-value connection 'lock)
+                ;; Shorter than the setup header.
+                (> 8 (length cache)))
+      (setf (slot-value connection 'lock) t)
+      (let ((data-len (+ 8 (* 4 (if xcb:lsb
+                                    (xcb:-unpack-u2-lsb cache 6)
+                                  (xcb:-unpack-u2 cache 6)))))
+            obj)
+        (when (>= (length cache) data-len)
+          (xcb:-log "Setup response: %s" cache)
+          (pcase (aref cache 0)
+            (0
+             ;; Connection failed.
+             (setq obj (make-instance 'xcb:SetupFailed))
+             (xcb:unmarshal obj cache)
+             (setq cache (substring cache data-len))
+             (error "[XELB] Connection failed: %s" (slot-value obj 'reason)))
+            (1
+             ;; Connection established.
+             (setf (slot-value connection 'message-cache) [])
+             (set-process-filter process #'xcb:-connection-filter)
+             (setq obj (make-instance 'xcb:Setup))
+             (xcb:unmarshal obj cache)
+             (setq cache (substring cache data-len))
+             (setf (slot-value connection 'setup-data) obj)
+             (setf (slot-value connection 'connected) t))
+            (2
+             ;; Authentication required.
+             (setq obj (make-instance 'xcb:SetupAuthenticate))
+             (xcb:unmarshal obj cache)
+             (setq cache (substring cache data-len))
+             (error "[XELB] Authentication not supported: %s"
+                    (slot-value obj 'reason)))
+            (x (error "Unrecognized setup status: %d" x)))))
+      (setf (slot-value connection 'lock) nil))))
+
+(cl-defmethod xcb:-convert-sequence ((obj xcb:connection) sequence16)
+  "Convert 16-bit sequence number SEQUENCE16 (read from a packet).
+
+The result would be 29 or 61 bits, depending on the machine."
+  (with-slots (request-sequence) obj
+    ;; Assume there are no more than #xFFFF requests sent since the
+    ;; request corresponding to this packet was made.  Because errors
+    ;; and replies are always read out in the process filter, this
+    ;; assumption is quite safe.
+    (let ((sequence (logior (logand request-sequence
+                                    xcb:-SEQUENCE-SEGMENT-MASK)
+                            sequence16)))
+      ;; `xcb:-cache-request' ensures sequence number never wraps.
+      (when (> sequence request-sequence)
+        (cl-decf sequence #x10000))
+      sequence)))
+
 (defun xcb:-connection-filter (process message)
   "Filter function for an X connection.
 
@@ -242,43 +289,9 @@ Concurrency is disabled as it breaks the orders of errors, replies and events."
          (cache (vconcat (slot-value connection 'message-cache) message))
          (cache-length (length cache)))
     (setf (slot-value connection 'message-cache) cache)
-    (catch 'return
-      ;; Queue message when locked
-      (when (slot-value connection 'lock)
-        (throw 'return 'lock))
+    (unless (slot-value connection 'lock)
       ;; Start parsing message
       (setf (slot-value connection 'lock) t)
-      ;; Connection setup
-      (unless (slot-value connection 'connected)
-        (when (<= 8 (length cache)) ;at least setup header is available
-          (let ((data-len (+ 8 (* 4 (funcall (if xcb:lsb #'xcb:-unpack-u2-lsb
-                                               #'xcb:-unpack-u2)
-                                             cache 6))))
-                obj)
-            (when (>= (length cache) data-len)
-              (xcb:-log "Setup response: %s" cache)
-              (pcase (aref cache 0)
-                (0                      ;failed
-                 (setq obj (make-instance 'xcb:SetupFailed))
-                 (xcb:unmarshal obj cache)
-                 (setq cache (substring cache data-len))
-                 (error "[XELB] Connection failed: %s"
-                        (slot-value obj 'reason)))
-                (1                      ;success
-                 (setq obj (make-instance 'xcb:Setup))
-                 (xcb:unmarshal obj cache)
-                 (setq cache (substring cache data-len))
-                 (setf (slot-value connection 'setup-data) obj)
-                 (setf (slot-value connection 'connected) t))
-                (2                      ;authentication
-                 (setq obj (make-instance 'xcb:SetupAuthenticate))
-                 (xcb:unmarshal obj cache)
-                 (setq cache (substring cache data-len))
-                 (error "[XELB] Authentication not supported: %s"
-                        (slot-value obj 'reason)))
-                (x (error "Unrecognized setup status: %d" x))))))
-        (setf (slot-value connection 'lock) nil)
-        (throw 'return 'setup))
       ;; Process error/reply/event
       (catch 'break
         (while (<= 32 (length cache))
@@ -290,6 +303,7 @@ Concurrency is disabled as it breaks the orders of errors, replies and events."
                                       cache 2))
                    (plist (slot-value connection 'error-plist))
                    struct)
+               (setq sequence (xcb:-convert-sequence connection sequence))
                (when (plist-member plist sequence)
                  (setq struct (plist-get plist sequence))
                  (setf (slot-value connection 'error-plist)
@@ -298,7 +312,7 @@ Concurrency is disabled as it breaks the orders of errors, replies and events."
                                           ,(substring cache 0 32))
                                         struct))))
                (setq cache (substring cache 32))
-               (setf (slot-value connection 'error-sequence) sequence)))
+               (setf (slot-value connection 'last-seen-sequence) sequence)))
             (1                          ;reply
              (let* ((reply-words (funcall (if xcb:lsb #'xcb:-unpack-u4-lsb
                                             #'xcb:-unpack-u4)
@@ -310,7 +324,8 @@ Concurrency is disabled as it breaks the orders of errors, replies and events."
                (xcb:-log "Reply received: %s" (substring cache 0 reply-length))
                (setq sequence (funcall (if xcb:lsb #'xcb:-unpack-u2-lsb
                                          #'xcb:-unpack-u2)
-                                       cache 2))
+                                       cache 2)
+                     sequence (xcb:-convert-sequence connection sequence))
                (setq plist (slot-value connection 'reply-plist))
                (setq struct (plist-get plist sequence))
                (when struct
@@ -325,22 +340,45 @@ Concurrency is disabled as it breaks the orders of errors, replies and events."
                                     `(,(car struct) ,@(cdr struct)
                                       ,(substring cache 0 reply-length))))))
                (setq cache (substring cache reply-length))
-               (setf (slot-value connection 'reply-sequence) sequence)))
+               (setf (slot-value connection 'last-seen-sequence) sequence)))
             (x                          ;event
-             (xcb:-log "Event received: %s" (substring cache 0 32))
-             (let (synthetic listener)
+             (let (synthetic listener event-length)
                (when (/= 0 (logand x #x80)) ;synthetic event
                  (setq synthetic t
                        x (logand x #x7f))) ;low 7 bits is the event number
                (setq listener
                      (plist-get (slot-value connection 'event-plist) x))
+               (pcase listener
+                 (`xge
+                  (setq event-length (funcall (if xcb:lsb
+                                                  #'xcb:-unpack-u4-lsb
+                                                #'xcb:-unpack-u4)
+                                              cache 4))
+                  (when (< (length cache) event-length)
+                    ;; Too short.
+                    (throw 'break nil))
+                  (setq listener
+                        (lax-plist-get (slot-value connection 'event-plist)
+                                       (vector (aref cache 1)
+                                               (funcall
+                                                (if xcb:lsb
+                                                    #'xcb:-unpack-u2-lsb
+                                                  #'xcb:-unpack-u2)
+                                                cache 8)))))
+                 (`xkb
+                  (setq listener
+                        (lax-plist-get (slot-value connection 'event-plist)
+                                       (vector (aref cache 1))))))
                (when listener
                  (with-slots (event-queue) connection
                    (setf event-queue (nconc event-queue
                                             `([,listener
                                                ,(substring cache 0 32)
-                                               ,synthetic]))))))
-             (setq cache (substring cache 32))))))
+                                               ,synthetic])))))
+               (unless event-length
+                 (setq event-length 32))
+               (xcb:-log "Event received: %s" (substring cache 0 event-length))
+               (setq cache (substring cache event-length)))))))
       (setf (slot-value connection 'lock) nil))
     (unless (slot-value connection 'lock)
       (with-slots (message-cache) connection
@@ -364,12 +402,14 @@ Concurrency is disabled as it breaks the orders of errors, replies and events."
 
 (cl-defmethod xcb:disconnect ((obj xcb:connection))
   "Disconnect from X server."
-  (delete-process (slot-value obj 'process))
-  ;; Reset every slot to its default value
-  (let ((slots (eieio-class-slots 'xcb:connection)))
-    (dolist (slot slots)
-      (setf (slot-value obj (eieio-slot-descriptor-name slot))
-            (eieio-oref-default obj (eieio-slot-descriptor-name slot))))))
+  (when (slot-value obj 'connected)
+    (xcb:flush obj)
+    (delete-process (slot-value obj 'process))
+    ;; Reset every slot to its default value
+    (let ((slots (eieio-class-slots 'xcb:connection)))
+      (dolist (slot slots)
+        (setf (slot-value obj (eieio-slot-descriptor-name slot))
+              (eieio-oref-default obj (eieio-slot-descriptor-name slot)))))))
 
 ;;;; Other routines
 
@@ -386,15 +426,26 @@ Concurrency is disabled as it breaks the orders of errors, replies and events."
   (slot-value (xcb:get-setup obj) 'maximum-request-length))
 
 (cl-defmethod xcb:+event ((obj xcb:connection) event listener)
-  "attach EVENT LISTENER
+  "Attach function LISTENER to event EVENT.
 
 Note that event listeners attached this way are shared with the super- and sub-
 classes of EVENT (since they have the same event number)."
   (let* ((event-number (xcb:-error-or-event-class->number obj event))
          (plist (slot-value obj 'event-plist))
-         (listeners (plist-get plist event-number)))
+         key listeners)
+    (when (consp event-number)
+      (setq key (car event-number)
+            event-number (cdr event-number)
+            listeners (plist-get plist key))
+      ;; Add a placeholder.
+      (setf (slot-value obj 'event-plist)
+            (plist-put plist key
+                       (if (child-of-class-p event 'xcb:-generic-event)
+                           'xge 'xkb))))
+    (setq listeners (lax-plist-get plist event-number))
     (setf (slot-value obj 'event-plist)
-          (plist-put plist event-number (append listeners (list listener))))))
+          (lax-plist-put plist event-number (append listeners
+                                                    (list listener))))))
 
 (cl-defmethod xcb:flush ((obj xcb:connection))
   "Flush request data to X server."
@@ -491,22 +542,26 @@ classes of EVENT (since they have the same event number)."
              (+ (length msg) (length cache))) ;flush on cache full
       (xcb:flush obj)
       (setq cache []))
-    (with-slots (request-cache request-sequence) obj
+    (with-slots (request-cache request-sequence last-seen-sequence) obj
+      (when (>= request-sequence most-positive-fixnum)
+        ;; Force wrapping the sequence number.
+        (xcb:aux:sync obj)
+        (setf request-sequence 0
+              last-seen-sequence 0))
       (setf request-cache (vconcat cache msg)
             request-sequence (1+ request-sequence))
       (xcb:-log "Cache request #%d: %s" request-sequence request)
       request-sequence)))
 
 (cl-defmethod xcb:-+request ((obj xcb:connection) request)
-  (let* ((sequence (xcb:-cache-request obj request))
-         (sequence-lsw (logand #xFFFF sequence))
-         (class (eieio-object-class request)))
+  (let ((sequence (xcb:-cache-request obj request))
+        (class (eieio-object-class request)))
     (when (fboundp (xcb:-request-class->reply-class class))
       ;; This request has a reply
       (setf (slot-value obj 'reply-plist) ;require reply
-            (plist-put (slot-value obj 'reply-plist) sequence-lsw class))
+            (plist-put (slot-value obj 'reply-plist) sequence class))
       (setf (slot-value obj 'error-plist) ;require error
-            (plist-put (slot-value obj 'error-plist) sequence-lsw nil)))
+            (plist-put (slot-value obj 'error-plist) sequence nil)))
     sequence))
 
 (defmacro xcb:+request (obj request)
@@ -521,10 +576,9 @@ Otherwise no error will ever be reported."
   (when (fboundp
          (xcb:-request-class->reply-class (eieio-object-class request)))
     (error "This method shall not be called with request that has a reply"))
-  (let* ((sequence (xcb:-cache-request obj request))
-         (sequence-lsw (logand #xFFFF sequence)))
+  (let ((sequence (xcb:-cache-request obj request)))
     (setf (slot-value obj 'error-plist)
-          (plist-put (slot-value obj 'error-plist) sequence-lsw nil))
+          (plist-put (slot-value obj 'error-plist) sequence nil))
     sequence))
 
 (defmacro xcb:+request-checked (obj request)
@@ -536,11 +590,10 @@ Otherwise no error will ever be reported."
   (unless (fboundp
            (xcb:-request-class->reply-class (eieio-object-class request)))
     (error "This method shall not be called with request that has no reply"))
-  (let* ((sequence (xcb:-cache-request obj request))
-         (sequence-lsw (logand #xFFFF sequence)))
+  (let ((sequence (xcb:-cache-request obj request)))
     (setf (slot-value obj 'reply-plist)
           (plist-put (slot-value obj 'reply-plist)
-                     sequence-lsw (eieio-object-class request)))
+                     sequence (eieio-object-class request)))
     sequence))
 
 (defmacro xcb:+request-unchecked (obj request)
@@ -549,27 +602,20 @@ Otherwise no error will ever be reported."
   `(xcb:-+request-unchecked ,obj ,request))
 
 (cl-defmethod xcb:-+reply ((obj xcb:connection) sequence &optional multiple)
-  (setq sequence (logand #xFFFF sequence)) ;only the LSW is used
   (unless (plist-member (slot-value obj 'reply-plist) sequence)
     (error "This method is intended for requests with replies"))
   (xcb:flush obj)                      ;or we may have to wait forever
   (if multiple
       ;; Multiple replies
-      (when (and (<= 0 (xcb:-sequence-cmp16 sequence
-                                            (slot-value obj 'reply-sequence)))
-                 (<= 0 (xcb:-sequence-cmp16 sequence
-                                            (slot-value obj 'error-sequence))))
-        (xcb:aux:sync obj))
+      (xcb:aux:sync obj)
     ;; Single reply
     (let ((process (slot-value obj 'process)))
       ;; Wait until the request processed
       (cl-incf (slot-value obj 'event-lock))
       (with-timeout (xcb:connection-timeout
                      (warn "[XELB] Retrieve reply timeout"))
-        (while (and (< 0 (xcb:-sequence-cmp16
-                          sequence (slot-value obj 'reply-sequence)))
-                    (< 0 (xcb:-sequence-cmp16
-                          sequence (slot-value obj 'error-sequence))))
+        (while (and (> sequence (slot-value obj 'last-seen-sequence))
+                    (<= sequence (slot-value obj 'request-sequence)))
           (accept-process-output process 1 nil 1)))
       (cl-decf (slot-value obj 'event-lock))))
   (let* ((reply-plist (slot-value obj 'reply-plist))
@@ -583,14 +629,12 @@ Otherwise no error will ever be reported."
       (if multiple
           ;; Multiple replies
           (dolist (i (cdr reply-data))
-            (setq reply (make-instance class-name
-                                       :length (/ (- (length i) 32) 4)))
+            (setq reply (make-instance class-name))
             (xcb:unmarshal reply i)
             (setq replies (nconc replies (list reply))))
         ;; Single reply
         (setq reply-data (cadr reply-data)
-              replies (make-instance class-name
-                                     :length (/ (- (length reply-data) 32) 4)))
+              replies (make-instance class-name))
         (xcb:unmarshal replies reply-data)))
     (setq errors
           (mapcar (lambda (i)
@@ -615,7 +659,6 @@ MULTIPLE value, or some replies may be lost!"
   `(xcb:-+reply ,obj ,sequence ,multiple))
 
 (cl-defmethod xcb:-request-check ((obj xcb:connection) sequence)
-  (setq sequence (logand #xFFFF sequence)) ;only the LSW is used
   (when (plist-member (slot-value obj 'reply-plist) sequence)
     (error "This method is intended for requests with no reply"))
   (xcb:flush obj)                      ;or we may have to wait forever
@@ -623,7 +666,7 @@ MULTIPLE value, or some replies may be lost!"
         error-obj tmp)
     (unless (plist-member error-plist sequence)
       (error "This method shall be called after `xcb:+request-checked'"))
-    (when (< 0 (xcb:-sequence-cmp16 sequence (slot-value obj 'error-sequence)))
+    (when (> sequence (slot-value obj 'last-seen-sequence))
       (xcb:aux:sync obj))         ;wait until the request is processed
     (setq error-obj
           (mapcar (lambda (i)
@@ -677,45 +720,85 @@ MULTIPLE value, or some replies may be lost!"
   "Force sync with X server.
 
 Sync by sending a GetInputFocus request and waiting until it's processed."
-  (let* ((sequence (xcb:-cache-request obj (make-instance 'xcb:GetInputFocus)))
-         (sequence-lsw (logand #xFFFF sequence))
-         (process (slot-value obj 'process)))
+  (let ((sequence (xcb:-cache-request obj (make-instance 'xcb:GetInputFocus)))
+        (process (slot-value obj 'process)))
     (xcb:flush obj)
     ;; Wait until request processed
     (cl-incf (slot-value obj 'event-lock))
     (with-timeout (xcb:connection-timeout (warn "[XELB] Sync timeout"))
-      (while (< 0 (xcb:-sequence-cmp16 sequence-lsw
-                                       (slot-value obj 'reply-sequence)))
+      (while (and (> sequence (slot-value obj 'last-seen-sequence))
+                  ;; In case the sequence number has been wrapped.
+                  (<= sequence (slot-value obj 'request-sequence)))
         (accept-process-output process 1 nil 1)))
-    (cl-decf (slot-value obj 'event-lock))))
+    (cl-decf (slot-value obj 'event-lock))
+    ;; Discard any reply or error.
+    (cl-remf (slot-value obj 'reply-plist) sequence)
+    (cl-remf (slot-value obj 'error-plist) sequence)))
 
 (cl-defmethod xcb:-error-or-event-class->number ((obj xcb:connection) class)
-  "Return the error/event number of a error/event class CLASS."
+  "Return the error/event number of a error/event class CLASS.
+
+If CLASS is a generic event, return (XGE-CODE . [EXTENSION EVTYPE]);
+Or if it's an XKB event, return (XKB-EVENT-CODE [XKB-CODE])."
   (unless (symbolp class) (setq class (eieio-class-name class)))
-  (let* ((is-error (child-of-class-p class 'xcb:-error))
-         (prefix (replace-regexp-in-string ":[^:]+$" ":" (symbol-name class)))
-         (first (when (string= prefix "xcb:") 0))
-         (alist (intern-soft (concat prefix (if is-error
-                                                "error-number-class-alist"
-                                              "event-number-class-alist"))))
-         result parents)
-    (unless first
-      (setq first (cdr (assoc (intern (substring prefix 0 -1))
-                              (slot-value obj
-                                          (if is-error
-                                              'extension-first-error-alist
-                                            'extension-first-event-alist))))))
-    (when alist
-      (setq result (+ (or first 0) (car (rassoc class (symbol-value alist))))))
-    (if result
-        result
+  (let ((prefix (replace-regexp-in-string ":[^:]+$" ":" (symbol-name class)))
+        first-code alist result parents)
+    (cond
+     ((child-of-class-p class 'xcb:-error)
+      ;; Error.
+      (if (string= prefix "xcb:")
+          (setq first-code 0
+                alist xcb:error-number-class-alist)
+        (setq first-code
+              (cdr (assq (intern (substring prefix 0 -1))
+                         (slot-value obj
+                                     'extension-first-error-alist)))
+              alist (symbol-value
+                     (intern-soft (concat prefix
+                                          "error-number-class-alist")))))
+      (setq result (car (rassq class alist)))
+      (when result
+        (setq result (+ first-code result))))
+     ((child-of-class-p class 'xcb:-generic-event)
+      ;; Generic event.
+      (setq alist (symbol-value
+                   (intern-soft (concat prefix "xge-number-class-alist")))
+            result (plist-get (slot-value obj 'extension-opcode-plist) class))
+      ;; Ensure the extension has been initialized.
+      (when result
+        (setq result `(35 . [,result ,(cdr (rassq class alist))]))))
+     ((string= prefix "xcb:xkb:")
+      ;; XKB event.
+      (eval-and-compile (require 'xcb-xkb))
+      ;; XKB uses a single event code for all events.
+      (setq result (cdr (assq 'xcb:xkb
+                              (slot-value obj 'extension-first-event-alist))))
+      ;; Ensure the XKB extension has been initialized.
+      (when result
+        (setq alist xcb:xkb:event-number-class-alist
+              result `(,result . [,(car (rassq class alist))]))))
+     (t
+      ;; Other event.
+      (if (string= prefix "xcb:")
+          (setq first-code 0
+                alist xcb:event-number-class-alist)
+        (setq first-code
+              (cdr (assq (intern (substring prefix 0 -1))
+                         (slot-value obj 'extension-first-event-alist)))
+              alist (symbol-value
+                     (intern-soft (concat prefix
+                                          "event-number-class-alist")))))
+      (setq result (car (rassq class alist)))
+      (when result
+        (setq result (+ first-code result)))))
+    (unless result
       ;; Fallback to use the error/event number of one superclass. Thus if the
       ;; error/event number of a subclass differs from that of its parent, it
       ;; must be explicitly pointed out.
       (setq parents (eieio-class-parents class))
       (while (and parents (not result))
-        (setq result (xcb:-error-or-event-class->number obj (pop parents))))
-      result)))
+        (setq result (xcb:-error-or-event-class->number obj (pop parents)))))
+    result))
 
 (cl-defmethod xcb:-event-number->class ((obj xcb:connection) number)
   "Return the event class that has the event number NUMBER.
